"""Finding candidate pairs from multiple datasets."""

import array as _array
import collections as _collections
import heapq as _heapq
import itertools as _itertools
import typing as _typing

import anonlink.typechecking as _typechecking

_Block = _typing.Tuple[_typing.List[int], ...]
_CandidatePair = _typing.Tuple[float, int, int, int, int]
_CandidatePairIterable = _typing.Iterable[_CandidatePair]


def _to_candidate_pairs(
    sims: _typing.Iterable[float],
    rec_is0: _typing.Iterable[int],
    rec_is1: _typing.Iterable[int],
    i0: int,
    i1: int,
    block: _typing.Sequence[_typing.Sequence[int]]
) -> _CandidatePairIterable:
    return ((sim, i0, i1, block[i0][rec_i0], block[i1][rec_i1])
            for sim, rec_i0, rec_i1 in zip(sims, rec_is0, rec_is1))


def _block_similarities(
    block: _Block,
    datasets: _typing.Sequence[_typechecking.Dataset],
    similarity_f: _typechecking.SimilarityFunction,
    threshold: float,
    k: _typing.Optional[int]
) -> _typing.Iterable[_CandidatePairIterable]:
    for i0, i1 in _itertools.combinations(range(len(block)), 2):
        recs_dset0 = tuple(map(datasets[i0].__getitem__, block[i0]))
        recs_dset1 = tuple(map(datasets[i1].__getitem__, block[i1]))
        if len(recs_dset0) > 0 and len(recs_dset1) > 0:
            sims, (rec_is0, rec_is1) = similarity_f((recs_dset0, recs_dset1), threshold, k=k)
            yield _to_candidate_pairs(sims, rec_is0, rec_is1, i0, i1, block)


def _enforce_k(
    similarities: _CandidatePairIterable,
    k: int
) -> _CandidatePairIterable:
    candidates_counter: _typing.Counter[_typing.Tuple[int, int, int]] \
        = _collections.Counter()
    for similarity in similarities:
        _, dset_i0, dset_i1, rec_i0, rec_i1 = similarity
        # At most k candidate pairs per record per dataset pair.
        i0 = dset_i0, dset_i1, rec_i1
        i1 = dset_i1, dset_i0, rec_i0
        candidates_counter[i0] += 1
        candidates_counter[i1] += 1
        if candidates_counter[i0] <= k and candidates_counter[i1] <= k:
            yield similarity


def _merge_similarities(
    similarities: _typing.Iterable[_CandidatePairIterable],
    k: _typing.Optional[int]
) -> _typechecking.CandidatePairs:
    # Merge multiple sorted sequences
    sorted_similarities = _heapq.merge(*similarities,
                                       key=lambda x: (-x[0], *x[1:]))
    
    # One record can be in multiple blocks. Remove duplicates.
    deduplicated_similarities = (
        k for k, _ in _itertools.groupby(sorted_similarities))
    
    if k is None:
        k_enforced_similarities: _CandidatePairIterable \
            = deduplicated_similarities
    else:
        k_enforced_similarities = _enforce_k(deduplicated_similarities, k)

    # Assume all arrays are the same type.
    # Future: this may require changing.
    sims: _typechecking.FloatArrayType = _array.array('d')
    dset_is0: _typechecking.IntArrayType = _array.array('I')
    dset_is1: _typechecking.IntArrayType = _array.array('I')
    rec_is0: _typechecking.IntArrayType = _array.array('I')
    rec_is1: _typechecking.IntArrayType = _array.array('I')
    for sim, dset_i0, dset_i1, rec_i0, rec_i1 in k_enforced_similarities:
        sims.append(sim)
        dset_is0.append(dset_i0)
        dset_is1.append(dset_i1)
        rec_is0.append(rec_i0)
        rec_is1.append(rec_i1)
    return sims, (dset_is0, dset_is1), (rec_is0, rec_is1)


def find_candidate_pairs(
    datasets: _typing.Sequence[_typechecking.Dataset],
    similarity_f: _typechecking.SimilarityFunction,
    threshold: float,
    k: _typing.Optional[int] = None,
    blocking_f: _typing.Optional[_typechecking.BlockingFunction] = None
) -> _typechecking.CandidatePairs:
    """Find candidate pairs from multiple datasets. Optional blocking.

    :param datasets: A sequence of datasets. Each dataset is a sequence
        of hashes.
    :param similarity_f: A function that computes a similarity matrix
        between two sequences of hashes and finds candidates above the
        threshold.
    :param threshold: The similarity threshold. We accept pairs that
        have similarity of at least this value.
    :param k: Only permit this many candidate pairs per dataset pair per
        record. Set to `None` to permit all pairs above with similarity
        at least `threshold`.
    :param blocking_f: A function returning all block IDs for a record.
        Two records are compared iff they have at least one block ID in
        common. Support for this is experimental and subject to change.

    :return: A 3-tuple `(similarity, dataset_i, record_i)`. `dataset_i`
        and `record_i` are sequences of sequences. Every sequence in
        `dataset_i` has the same length as `similarity`; also, every
        sequence in `record_i` has the same length as `similarity`.
        Currently `dataset_i` and `record_i` have length 2, but this may
        be changed in the future.

        Every valid index `i` corresponds to one candidate match.
        `dataset[0][i]` is the index of the dataset of the first record
        in the pair; `record[0][i]` is this record's index in its
        dataset. `dataset_[1][i]` is the index of the dataset of the
        second record in the pair; `record_[1][i]` is this record's
        index in its dataset. `similarity[i]` is the pair's similarity;
        this value will be greater than `threshold`.
    """
    if blocking_f is None:
        # Dummy blocking function.
        def blocking_f(dataset_index, record_index, hash_):
            return None,
    assert blocking_f is not None  # This is for Mypy.

    blocks: _typing.DefaultDict[_typing.Hashable, _Block] \
        = _collections.defaultdict(lambda: tuple([] for _ in datasets))
    for i, dataset in enumerate(datasets):
        for j, record in enumerate(dataset):
            for block_id in blocking_f(i, j, record):
                blocks[block_id][i].append(j)

    similarities = tuple(_itertools.chain.from_iterable(
        map(_block_similarities,
            blocks.values(),
            _itertools.repeat(datasets),
            _itertools.repeat(similarity_f),
            _itertools.repeat(threshold),
            _itertools.repeat(k))))

    return _merge_similarities(similarities, k)



"""Solvers.

Solvers accept candidate pairs and return a concrete matching. In other
words, they accept a similarity for every candidate pair and turn it
into a boolean for every candidate pair.
"""

import typing as _typing

import anonlink.typechecking as _typechecking

from anonlink.solving._multiparty_solving_python import (
    greedy_solve_python, probabilistic_greedy_solve_python)
try:
    from anonlink.solving._multiparty_solving import (
        greedy_solve_native, probabilistic_greedy_solve_native)
except ImportError:
    greedy_solve = greedy_solve_python
    probabilistic_greedy_solve = probabilistic_greedy_solve_python
else:
    greedy_solve = greedy_solve_native
    probabilistic_greedy_solve = probabilistic_greedy_solve_native


def pairs_from_groups(
    groups: _typechecking.MatchGroups
) -> _typing.Iterable[_typing.Tuple[int, int]]:
    """Make an iterable of pairs (i, j) from an iterable of groups.

    The set of pairs representation is often more convenient for
    bipartite matching problems.

    To make a mapping, use ``dict(pairs_from_groups(groups))``.

    :param groups: An iterable of groups.
    :return: An iterable of pairs (i, j), where i is the record index in
        the first dataset, and j is the record index in the second
        dataset.
    """
    group: _typing.List[_typing.Tuple[int, int]]
    for group in map(sorted, groups):  # type: ignore
        if len(group) == 2:
            (dset_i0, rec_i0), (dset_i1, rec_i1) = group
            if dset_i0 == 0 and dset_i1 == 1:
                yield rec_i0, rec_i1
                continue
        raise ValueError('non-bipartite problems are unsupported')



from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.unordered_set cimport unordered_set


cdef extern from "_multiparty_solving_inner.h":
    cdef struct Record:
        unsigned int dset_i
        unsigned int rec_i

    ctypedef vector[Record] Group

    unordered_set[Group *] greedy_solve_inner(
        unsigned int[],
        unsigned int[],
        unsigned int[],
        unsigned int[],
        size_t,
        double,
        bool
    ) nogil except +
    # `except +` asks Cython to propagate C++ exceptions to Python land



"""Similarity functions.

These functions take datasets of records and compute the records'
similarities. A threshold must be passed: only record pairs with
similarity of at least this threshold are returned. We call these the
candidate pairs.

Currently, the Dice Coefficient and the Simple Matching Coefficient are
implemented. These work on binary strings. However, other similarity
functions are possible as well.
"""

from anonlink.similarities._dice_python import dice_coefficient_python
from anonlink.similarities._smc import (hamming_similarity,
                                        simple_matching_coefficient)

try:
    from anonlink.similarities._dice_x86 import dice_coefficient_accelerated
except ImportError:
    # Alias that works even if the import fails, but always points to
    # the fastest implementation.
    dice_coefficient = dice_coefficient_python
else:
    dice_coefficient = dice_coefficient_accelerated



from libcpp cimport bool
from libcpp.vector cimport vector

from cython.operator cimport dereference as deref

from anonlink.solving._multiparty_solving_inner cimport (
    Record, Group, greedy_solve_inner)


def probabilistic_greedy_solve_native(
    candidates,
    *,
    merge_threshold=.5,
    deduplicated=True
):
    """Select matches using the probabilistic greedy algorithm.

    This solver is similar to `greedy_solve_native`, but may match
    records that do not have corresponding candidate pairs when there is
    enough other evidence that they are a match. This usually results in
    better accuracy.

    We assign each record to exactly one 'group' of records that are
    pairwise matched together. We iterate over `candidates` in order of
    decreasing similarity. When we encounter a pair of records (let s be
    their similarity) that do not already belong to the same group, we
    merge their groups iff (a) the proportion of pairs that are
    permitted to be matched in the groups' records is at least
    `merge_threshold`, and (b) the proportion of pairs whose similarity
    is above s in the groups' records is at least `merge_threshold`. The
    latter requirement is only for tie-breaking: we prefer groups whose
    minimum pairwise similarity is highest. Any group merge rejected due
    to requirement (b) will be revisited later with that requirement
    relaxed.

    :param tuple candidates: Candidates, as returned by
        `find_candidates`.
    :param float merge_threshold: Proportion of agreement between groups
        required to merge them. In (0, 1].
            When considering whether to merge two groups, we compute the
        groups' Cartesian product. We count the pairs in that product
        whose similarity is above the threshold. We merge the groups if
        our count is at least `merge_threshold` of the total number of
        pairs between the two groups.
            For example, suppose that we have two groups of records: 
        {A, B} and {C, D}. Let A-C, A-D and B-C be candidate pairs.
        Setting `merge_threshold` to any value not greater than .75 will
        permit the two groups to be merged since they share three
        candidate pairs out of all four pairs. Even though the
        similarity of B-D is below the threshold, there is enough
        evidence in the other pairs to conclude that it must be a match.
        The `merge_threshold` can be thought of as the inverse of the
        weight we place on that evidence.
            Higher values increase precision, but decrease recall.
            A `merge_threshold` lower than one permits two records to be
        matched without having a corresponding candidate pair. Set it to
        1.0 to ensure that two records will only be matched if their
        similarity is above the similarity threshold.
    :param bool deduplicated: When True, two records that belong to the
        same dataset will never be in the same group. Default True.

    :return: An sequence of groups. Each group is an sequence of
        records. Two records are in the same group iff they represent
        the same entity. Here, a record is a two-tuple of dataset index
        and record index.
    """
    sims_arr, dset_is_arrs, rec_is_arrs = candidates
    if len(dset_is_arrs) != len(rec_is_arrs):
        raise ValueError('inconsistent shape of index arrays')
    if len(dset_is_arrs) != 2:
        raise NotImplementedError('only binary solving is supported')
    if merge_threshold < 0 or merge_threshold > 1:
        raise ValueError('merge_threshold must be between 0 and 1')

    dset_is0_arr, dset_is1_arr = dset_is_arrs
    rec_is0_arr, rec_is1_arr = rec_is_arrs
    cdef size_t n = <size_t>len(sims_arr)
    if not (n
            == len(dset_is0_arr) == len(dset_is1_arr)
            == len(rec_is0_arr) == len(rec_is1_arr)):
        raise ValueError('inconsistent shape of index arrays')

    # NB: [::1] makes sure that the array is C-contiguous
    cdef unsigned int[::1] dset_is0 = dset_is0_arr
    cdef unsigned int[::1] dset_is1 = dset_is1_arr
    cdef unsigned int[::1] rec_is0 = rec_is0_arr
    cdef unsigned int[::1] rec_is1 = rec_is1_arr

    cdef double merge_threshold_double = merge_threshold
    cdef bool deduplicated_bool = deduplicated

    if n:  # Prevent dereferencing empty arrays.
        with nogil:
            cpp_result = greedy_solve_inner(
                &dset_is0[0],
                &dset_is1[0],
                &rec_is0[0],
                &rec_is1[0],
                n,
                merge_threshold_double,
                deduplicated_bool)

        # Save groups of size > 1.
        result = tuple(tuple((record.dset_i, record.rec_i)
                             for record in deref(group))
                       for group in cpp_result if deref(group).size() > 1)
        # Free all the memory!
        for group in cpp_result:
            del group
        return result
    else:
        return ()


def greedy_solve_native(candidates):
    """Select matches from candidate pairs using the greedy algorithm.

    We assign each record to exactly one 'group' of records that are
    pairwise matched together. We iterate over `candidates` in order of
    decreasing similarity. When we encounter a pair of records (let s be
    their similarity) that do not already belong to the same group, we
    merge their groups iff (a) every pair of records is permitted to be
    matched, and (b) the similarity of every pair of records is above s.
    The latter requirement is only for tie-breaking: we prefer groups
    whose minimum pairwise similarity is highest. Any group merge
    rejected due to requirement (b) will be revisited later with that
    requirement relaxed.

    :param tuple candidates: Candidates, as returned by
        `find_candidates`.

    :return: An sequence of groups. Each group is an sequence of
        records. Two records are in the same group iff they represent
        the same entity. Here, a record is a two-tuple of dataset index
        and record index.
    """
    return probabilistic_greedy_solve_native(
        candidates, merge_threshold=1.0, deduplicated=False)



"""Types for typechecking. Aimed at Mypy."""

import array as _array
import typing as _typing

import mypy_extensions as _mypy_extensions

# This is for parametrisation only. You shouldn't need to import it.
Record = _typing.TypeVar('Record')

Dataset = _typing.Sequence[Record]

if _typing.TYPE_CHECKING:
    FloatArrayType = _array.array[float]
    IntArrayType = _array.array[int]
    CharArrayType = _array.array[int]
else:
    FloatArrayType = _array.array
    IntArrayType = _array.array
    CharArrayType = _array.array

CandidatePairs = _typing.Tuple[FloatArrayType,
                       _typing.Tuple[IntArrayType, ...],
                       _typing.Tuple[IntArrayType, ...]]

BlockingFunction = _typing.Callable[
    [int, int, Record],
    _typing.Iterable[_typing.Hashable]]

SimilarityFunction = _typing.Callable[
    [_typing.Sequence[Dataset],
     float,
     _mypy_extensions.DefaultNamedArg(_typing.Optional[int], 'k')],
    _typing.Tuple[FloatArrayType, _typing.Sequence[IntArrayType]]]

DatasetChunkInfo = _mypy_extensions.TypedDict(
    'DatasetChunkInfo',
    {'datasetIndex': int,
     'range': _typing.List[int]})
ChunkInfo = _typing.List[DatasetChunkInfo]

DatasetAndRecordIndex = _typing.Tuple[int, int]
MatchGroups = _typing.Sequence[_typing.Sequence[DatasetAndRecordIndex]]



from array import array
from itertools import repeat
from typing import Iterable, Optional, Sequence, Tuple

from bitarray import bitarray

from anonlink.similarities._utils import (sort_similarities_inplace,
                                          to_bitarrays)
from anonlink.typechecking import FloatArrayType, IntArrayType

__all__ = ['dice_coefficient_python']


def dice_coefficient_python(
    datasets: Sequence[Sequence[bitarray]],
    threshold: float,
    k: Optional[int] = None
) -> Tuple[FloatArrayType, Tuple[IntArrayType, ...]]:
    """Find Dice coefficients of CLKs.

    This version is written in Python, so it does not rely on
    architecture-specific instructions. It may be slower than an
    accelerated version.

    The returned pairs are sorted in decreasing order of similarity,
    then in increasing order of the record index in the first dataset,
    and then in increasing order of the record index in the second
    dataset.

    :param datasets: A length 2 sequence of datasets. A dataset is a
        sequence of bitarrays.
    :param threshold: Pairs whose similarity is above this value may be
        a match.
    :param k: We only return the top k candidates for every record. Set
        to None to return all candidates.
    
    :return: A 2-tuple of similarity scores and indices. The similarity
        scores are an array of floating-point values. The indices are a
        2-tuple of arrays of integers.
    """
    n_datasets = len(datasets)
    if n_datasets < 2:
        raise ValueError(f'not enough datasets (expected 2, got {n_datasets})')
    elif n_datasets > 2:
        raise NotImplementedError(
            f'too many datasets (expected 2, got {n_datasets})')
    filters0, filters1 = datasets
    filters0 = to_bitarrays(filters0)
    filters1 = to_bitarrays(filters1)

    result_sims: FloatArrayType = array('d')
    result_indices0: IntArrayType = array('I')
    result_indices1: IntArrayType = array('I')

    if not filters0 or not filters1:
        # Empty result of the correct type.
        return result_sims, (result_indices0, result_indices1)

    f1_counts = tuple(f1.count() for f1 in filters1)

    for i, f0 in enumerate(filters0):
        f0_count = f0.count()
        if f0_count:
            coeffs: Iterable[float] = (
                2 * (f0 & f1).count() / (f0_count + f1_count)
                for f1, f1_count in zip(filters1, f1_counts))
        else:  # Avoid division by zero.
            coeffs = repeat(0., len(filters1))
        
        cands = filter(lambda c: c[1] >= threshold, enumerate(coeffs))
        top_k = sorted(cands, key=lambda x: -x[1])[:k]

        result_sims.extend(sim for _, sim in top_k)
        result_indices0.extend(repeat(i, len(top_k)))
        result_indices1.extend(j for j, _ in top_k)

    sort_similarities_inplace(result_sims, result_indices0, result_indices1)
    
    return result_sims, (result_indices0, result_indices1)



"""Helpers for concurrency."""

import array as _array
import itertools as _itertools
import math as _math
import typing as _typing

import numpy as _np

import anonlink.typechecking as _typechecking
from anonlink.candidate_generation import find_candidate_pairs

# Future: There may be better ways of chunking. Hamish suggests putting
# a better guarantee on the maximum size of a chunk. This may help with
# optimisation (e.g., set chunk size to be the size of a page,
# eliminating page faults).
# As the function currently makes no guarantees, any such changes would
# be backwards compatible.


def _split_points(size: int, chunks: int) -> _typing.Iterator[int]:
    chunk_size = size / chunks
    for i in range(chunks):
        yield round(i * chunk_size)
    yield size


def _chunks_1d(
    size: int,
    chunks: int
) -> _typing.Iterable[_typing.List[int]]:
    split_points = _split_points(size, chunks)
    a = next(split_points)
    for b in split_points:
        yield [a, b]
        a = b


def split_to_chunks(
    chunk_size_aim: float,
    *,
    # Keyword-only for forwards compatibility: this argument may not be
    # needed once we do blocking
    dataset_sizes: _typing.Sequence[int]
) -> _typing.Iterable[_typechecking.ChunkInfo]:
    """Split datasets into chunks for parallel processing.

    Resulting chunks are length 2 list of dictionaries. Each dictionary
    represents one data source for the chunk: it has a 'datasetIndex'
    key which maps to the index of the dataset as an integer, and it has
    a 'range' key mapping to the range of records within this dataset. A
    range is a length 2 list [a, b] representing range(a, b).

    Example: [{"datasetIndex": 2, "range": [3, 21]},
              {"datasetIndex": 4, "range": [18, 20]}]
    means that this chunk compares (0-indexed) datasets 2 and 4. We are
    looking at elements 3-20 (inclusive) of dataset 2 and elements 18
    and 19 of dataset 4.

    The chunks are always JSON serialisable.

    :param chunk_size_aim: Number of comparisons per chunk to aim for.
        This is a hint only. No promises.
    :param dataset_sizes: The sizes of the datasets to compare, as a
        sequence.

    :return: An iterable of chunks.
    """

    # int-like and float-like types such as np.int64 are welcome but are
    # not JSON-serialisable.
    chunk_size_aim_float = float(chunk_size_aim)
    dataset_sizes_int = map(int, dataset_sizes)
    for (i0, size0), (i1, size1) in _itertools.combinations(
            enumerate(dataset_sizes_int), 2):
        if not size0 or not size1:
            continue
        chunks0 = round(size0 / _math.sqrt(chunk_size_aim_float)) or 1
        chunk_size0 = size0 / chunks0
        # chunk_size0 is unlikely to be exactly sqrt(chunk_size_aim).
        # Adjust goal chunk size for the second dataset.
        chunks1 = round(size1 * chunk_size0 / chunk_size_aim_float) or 1
        for c0, c1 in _itertools.product(
                _chunks_1d(size0, chunks0), _chunks_1d(size1, chunks1)):
            yield [{'datasetIndex': i0, 'range': c0},
                   {'datasetIndex': i1, 'range': c1}]


def _get_dataset_indices(
    dataset_chunk: _typechecking.DatasetChunkInfo,
    size: int
) -> _typechecking.IntArrayType:
    index = dataset_chunk['datasetIndex']
    return _array.array('I', (index,)) * size


def _offset_record_indices_inplace(
    dataset_chunk: _typechecking.DatasetChunkInfo,
    rec_is: _typechecking.IntArrayType
) -> None:
    a, _ = dataset_chunk['range']
    np_rec_is = _np.frombuffer(rec_is, dtype=rec_is.typecode)
    np_rec_is += a


def _fill_int_array_inplace(
    int_array: _typechecking.IntArrayType,
    value: int
) -> None:
    np_buffer = _np.frombuffer(int_array, dtype=int_array.typecode)
    np_buffer.fill(value)


def process_chunk(
    chunk: _typechecking.ChunkInfo,
    datasets: _typing.Sequence[_typechecking.Dataset],
    similarity_f: _typechecking.SimilarityFunction,
    threshold: float,
    k: _typing.Optional[int] = None,
    blocking_f: _typing.Optional[_typechecking.BlockingFunction] = None
) -> _typechecking.CandidatePairs:
    """Find candidate pairs for the chunk.

    Calls the similarity function, offsets record indices by the
    required amount, and adds dataset index information.

    :param chunk: Chunk to process, as returned by `split_to_chunks`.
    :param datasets: A sequence of two datasets. Each dataset should
        contain as many records as required by `chunk`. It is up to you
        to extract the correct range from the larger dataset.
    :param similarity_f: A function that computes a similarity matrix
        between two sequences of hashes and finds candidates above the
        threshold.
    :param threshold: The similarity threshold. We accept pairs that
        have similarity of at least this value.
    :param k: Only permit this many candidate pairs per dataset pair per
        record. Set to `None` to permit all pairs above with similarity
        at least `threshold`.
    :param blocking_f: A function returning all block IDs for a record.
        Two records are compared iff they have at least one block ID in
        common. Support for this is experimental and subject to change.

    :return: A 3-tuple `(similarity, dataset_i, record_i)`. `dataset_i`
        and `record_i` are sequences of sequences. Every sequence in
        `dataset_i` has the same length as `similarity`; also, every
        sequence in `record_i` has the same length as `similarity`.
        Currently `dataset_i` and `record_i` have length 2, but this may
        be changed in the future.
            Every valid index `i` corresponds to one candidate match.
        `dataset[0][i]` is the index of the dataset of the first record
        in the pair; `record[0][i]` is this record's index in its
        dataset. `dataset_[1][i]` is the index of the dataset of the
        second record in the pair; `record_[1][i]` is this record's
        index in its dataset. `similarity[i]` is the pair's similarity;
        this value will be greater than `threshold`.
    """

    if len(chunk) != len(datasets):
        raise ValueError(
            f'number of datasets does not match chunk (expected {len(chunk)}, '
            f'got {len(datasets)})')
    for i, dataset_chunk in enumerate(chunk):
        if 'datasetIndex' not in dataset_chunk:
            raise ValueError(f"invalid chunk: expected value for "
                             f"'datasetIndex' at index {i}")
        if 'range' not in dataset_chunk:
            raise ValueError(f"invalid chunk: expected value for "
                             f"'range' at index {i}")
    for i, dataset_chunk, dataset_records in zip(
                _itertools.count(), chunk, datasets):
        a, b = dataset_chunk['range']
        if len(dataset_records) != b - a:
            raise ValueError(
                f'size of dataset at index {i} does not match chunk (expected '
                f'{b - a}, got {len(dataset_records)})')
    if len(chunk) != 2:
        raise NotImplementedError(
            f'only binary matching is currently supported '
            f'(chunk has {len(chunk)} datasets)')

    sims, (dset_is0, dset_is1), (rec_is0, rec_is1) = find_candidate_pairs(
        datasets,
        similarity_f,
        threshold,
        k=k,
        blocking_f=blocking_f
    )

    assert len(sims) == len(rec_is0) == len(rec_is1)

    _fill_int_array_inplace(dset_is0, chunk[0]['datasetIndex'])
    _offset_record_indices_inplace(chunk[0], rec_is0)

    _fill_int_array_inplace(dset_is1, chunk[1]['datasetIndex'])
    _offset_record_indices_inplace(chunk[1], rec_is1)

    return sims, (dset_is0, dset_is1), (rec_is0, rec_is1)



import itertools
import random
import operator
import os
from timeit import default_timer as timer

import bitarray
from clkhash.key_derivation import generate_key_lists
from clkhash.bloomfilter import stream_bloom_filters
from clkhash.randomnames import NameList

import anonlink


def generate_random_bitarrays(length=1024):
    while True:
        a = bitarray.bitarray()
        a.frombytes(os.urandom(length//8))
        yield a


def generate_random_clks(count, length=1024):
    return tuple(itertools.islice(generate_random_bitarrays(length), count))


some_filters = generate_random_clks(10000)


def print_comparison_header(headline):
    print()
    print(headline)
    print("Size 1 | Size 2 | Comparisons      | Total Time (s)          | Throughput")
    print("       |        |        (match %) | (comparisons / matching)|  (1e6 cmp/s)")
    print("-------+--------+------------------+-------------------------+-------------")


def compute_comparison_speed(n1, n2, threshold, k=None):
    """
    Using the greedy solver, how fast can hashes be computed using one core.
    """

    assert n1 != 0 and n2 != 0

    filters1 = [some_filters[random.randrange(0, 8000)] for _ in range(n1)]
    filters2 = [some_filters[random.randrange(2000, 10000)] for _ in range(n2)]

    start = timer()
    sparse_matrix = anonlink.candidate_generation.find_candidate_pairs(
        (filters1, filters2),
        anonlink.similarities.dice_coefficient,
        threshold,
        k=k
    )
    t1 = timer()
    anonlink.solving.greedy_solve(sparse_matrix)
    end = timer()

    similarity_time = t1 - start
    solver_time = end - t1
    elapsed_time = end - start

    inv_similarity_time = 1 / similarity_time if similarity_time > 0 else float('inf')
    inv_elapsed_time = 1 / elapsed_time if elapsed_time > 0 else float('inf')

    print("{:6d} | {:6d} | {:4d}e6  ({:5.2f}%) | {:6.3f}  ({:4.1f}% / {:4.1f}%) |  {:8.3f}".format(
        n1, n2, n1*n2 // 1000000,
        100.0*len(sparse_matrix[0])/(n1*n2),
        elapsed_time,
        100.0*similarity_time * inv_elapsed_time,
        100.0*solver_time * inv_elapsed_time,
        (n1*n2 / 1e6) * inv_similarity_time))
    return elapsed_time


def benchmark(size):

    print("Anonlink benchmark -- see README for explanation")
    print("------------------------------------------------")

    print(f"using '{anonlink.solving.greedy_solve.__name__}' as solver and "
          f"'{anonlink.similarities.dice_coefficient.__name__}' as similarity metric")

    possible_test_sizes = [
        1000, 2000, 3000, 4000,
        5000, 6000, 7000, 8000, 9000,
        10000,
        20000, 30000, 40000, 50000, 60000, 70000, 80000, 90000,
        1000000,
        2000000
    ]

    thld = 0.5
    print_comparison_header(f"Threshold: {thld}, All results returned")
    for test_size in possible_test_sizes:
        if test_size <= size:
            compute_comparison_speed(test_size, test_size, thld, k=None)

    size *= 5
    thld = 0.5
    print_comparison_header(f"Threshold: {thld}, Top 100 matches per record returned")
    for test_size in possible_test_sizes:
        if test_size <= size:
            compute_comparison_speed(test_size, test_size, thld, k=100)

    thld = 0.7
    print_comparison_header(f"Threshold: {thld}, All results returned")
    for test_size in possible_test_sizes:
        if test_size <= size:
            compute_comparison_speed(test_size, test_size, thld, k=None)

    print_comparison_header(f"Threshold: {thld}, Top 100 matches per record returned")
    for test_size in possible_test_sizes:
        if test_size <= size:
            compute_comparison_speed(test_size, test_size, thld, k=100)


if __name__ == '__main__':
    benchmark(4000)




import itertools
import typing as _typing

import numpy as _np

import anonlink.typechecking as _typechecking


def _similarities_as_nparray(candidate_pairs: _typechecking.CandidatePairs):
    sims, _, _ = candidate_pairs
    return _np.frombuffer(sims, dtype=sims.typecode)


def _check_bipartite(candidate_pairs: _typechecking.CandidatePairs) -> bool:
    _, (dset_is0, dset_is1), _ = candidate_pairs
    np_dset_is0 = _np.frombuffer(dset_is0, dtype=dset_is0.typecode)
    np_dset_is1 = _np.frombuffer(dset_is1, dtype=dset_is1.typecode)
    return (np_dset_is0 == 0).all() and (np_dset_is1 == 1).all()


def similarities_hist(candidate_pairs: _typechecking.CandidatePairs,
                      bins: int = 100):
    """Compute a histogram of the similarity scores in candidate pairs.

    This function is experimental and subject to change without warning.

    :param candidate_pairs: The candidate pairs.
    :param bins: An integer determining the number of bins to use.
        Default 100.

    :return: 2-tuple of (1) values of the histogram as an array length
        bins, and (2) the edges of the bins as an array length bins + 1.
    """
    return _np.histogram(_similarities_as_nparray(candidate_pairs),
                         bins=bins)


def _semiopen_hist_matches_nonmatches(
    candidate_pairs: _typechecking.CandidatePairs,
    steps: int = 100
):
    # Run the 2-party greedy solver.
    if not _check_bipartite(candidate_pairs):
        raise ValueError('only 2-party matching is supported')

    thresholds = _np.histogram_bin_edges(
        _similarities_as_nparray(candidate_pairs), bins=steps)
    thresholds_enumerate_iter = zip(
        range(thresholds.shape[0] - 1, -1, -1), thresholds[::-1])

    sims, _, (rec_is0, rec_is1) = candidate_pairs
    matched0: _typing.Set[int] = set()
    matched1: _typing.Set[int] = set()
    num_matches = _np.zeros_like(thresholds, dtype=int)
    num_nonmatches = _np.zeros_like(thresholds, dtype=int)
    try:
        i, threshold = next(thresholds_enumerate_iter)
    except StopIteration:
        return num_matches, num_nonmatches, thresholds
    for sim, rec_i0, rec_i1 in zip(sims, rec_is0, rec_is1):
        while sim < threshold:
            try:
                i, threshold = next(thresholds_enumerate_iter)
            except StopIteration:
                return num_matches, num_nonmatches, thresholds
        if rec_i0 not in matched0 and rec_i1 not in matched1:
            matched0.add(rec_i0)
            matched1.add(rec_i1)
            num_matches[i] += 1
        else:
            num_nonmatches[i] += 1
    return num_matches, num_nonmatches, thresholds


def matches_nonmatches_hist(
    candidate_pairs: _typechecking.CandidatePairs,
    bins: int = 100
):
    """Compute a histogram of possible matches and definite nonmatches.

    This function is experimental and subject to change without warning.

    We use the greedy solver to split the candidate pairs into possible
    matches and definite nonmatches. A possible match may or may not be
    accepted as a pair depending on the threshold chosen. A definite
    nonmatch will never be accepted, since one record in this pair has
    a more promising match with another record. We then make a histogram
    of possible matches and definite nonmatches.

    :param candidate_pairs: The candidate pairs.
    :param bins: An integer determining the number of bins to use.
        Default 100.

    :return: 3-tuple of (1) values of the histogram of the matches as an
        array of length bins, (2) values of the histogram of the
        nonmatches as an array of length bins, and (3) the edges of the
        bins as an array length bins + 1.
    """
    semiopen_hist = _semiopen_hist_matches_nonmatches(candidate_pairs, bins)
    num_matches, num_nonmatches, thresholds = semiopen_hist
    # Merge last and second last bins for consistency with np.histogram
    num_matches[-2] += num_matches[-1]
    num_matches = _np.array(num_matches[:-1])
    num_nonmatches[-2] += num_nonmatches[-1]
    num_nonmatches = _np.array(num_nonmatches[:-1])
    return num_matches, num_nonmatches, thresholds


def cumul_number_matches_vs_threshold(
    candidate_pairs: _typechecking.CandidatePairs,
    steps: int = 100
):
    """Compute the number of matches for each threshold.

    This function is experimental and subject to change without warning.

    We use the 2-party greedy solver to calculate the number of matches
    that would be returned if the candidate_pairs were found using a
    particular threshold. This function requires only a single pass of
    the data, so it is faster than simply running the greedy solver
    multiple times.

    :param candidate_pairs: The candidate pairs.
    :param steps: An integer determining the number of threshold steps
        to use. Default 100.

    :return: 2-tuple of (1) the number of matches for the threshold as
        an array length steps + 1, and (2) the thresholds as an array
        length steps + 1.
    """
    num_matches, _, thresholds = _semiopen_hist_matches_nonmatches(
        candidate_pairs, steps)
    num_matches_rev = num_matches[::-1]
    _np.cumsum(num_matches_rev, out=num_matches_rev)
    return num_matches, thresholds


def nonmatch_index_score(
    candidate_pairs: _typechecking.CandidatePairs,
    n: int = 1
) -> int:
    """Find the index of the ``n``th definite nonmatch.

    We use the 2-party greedy solver to split the candidate pairs into
    possible matches and definite nonmatches. The index (in decreasing
    order of similarity) of the ``n``th definite nonmatch is returned.

    Smaller values of ``n`` introduce more noise to the heuristic, but
    larger values of ``n`` provide less information.

    Raises ValueError if there are fewer than n definite nonmatches.

    :param candidate pairs: The candidate pairs. Must represent a
        bipartite problem.
    :param n: We return the index of the ``n``th definite nonmatch.
        Default 1.

    :return: The index of the ``n``th definite nonmatch if there are at
        least ``n`` definite nonmatches in the candidate pairs.
    """
    if not _check_bipartite(candidate_pairs):
        raise ValueError('only 2-party matching is supported')
    _, _, (rec_is0, rec_is1) = candidate_pairs

    matched0: _typing.Set[int] = set()
    matched1: _typing.Set[int] = set()
    nonmatches = 0
    for i, rec_i0, rec_i1 in zip(itertools.count(), rec_is0, rec_is1):
        if rec_i0 not in matched0 and rec_i1 not in matched1:
            matched0.add(rec_i0)
            matched1.add(rec_i1)
        else:
            nonmatches += 1
            if nonmatches == n:
                return i
    # Fewer than n definite nonmatches.
    raise ValueError('fewer than n definite nonmatches')




cdef extern from "dice.cpp":

    int match_one_against_many_dice_k_top(
            const char[] one,
            const char[] many,
            const unsigned int[] counts_many,
            int n,
            int keybytes,
            unsigned int k,
            double threshold,
            unsigned int[] indices,
            double[] scores
    ) nogil

    double dice_coeff(const char[] array1, const char[] array2, int array_bytes) nogil

    double popcount_arrays(
            unsigned int[] counts,
            const char[] arrays,
            unsigned int narrays,
            unsigned int array_bytes
    ) nogil




import collections
import heapq
from array import array
from typing import (Counter, DefaultDict, Iterable,
                    List, Optional, Sequence, Tuple)

from bitarray import bitarray

from anonlink import _deprecation
from anonlink.typechecking import Dataset, FloatArrayType, IntArrayType

__all__ = ['hamming_similarity', 'simple_matching_coefficient']

_Similarity = Tuple[float, int, int]

_deprecated_decorator = _deprecation.make_decorator('similarities')


def _smc_sim(clk1: bitarray, clk2: bitarray) -> float:
    n = len(clk1)
    return (n - (clk1 ^ clk2).count()) / n


def _smc_sims_gt_threshold(
    datasets: Sequence[Dataset[bitarray]],
    threshold: float
) -> Iterable[_Similarity]:
    dset0, dset1 = datasets
    for i0, clk0 in enumerate(dset0):
        for i1, clk1 in enumerate(dset1):
            sim = _smc_sim(clk0, clk1)
            if sim >= threshold:
                yield sim, i0, i1


def _smc_similarity_k(
    datasets: Sequence[Dataset[bitarray]],
    threshold: float,
    k: int
) -> Tuple[FloatArrayType, Tuple[IntArrayType, ...]]:
    # Dictionaries support sparsity, but these could also be list...
    candidates: Tuple[DefaultDict[int, List[_Similarity]], ...] \
        = tuple(collections.defaultdict(list) for _ in datasets)

    for sim, i0, i1 in _smc_sims_gt_threshold(datasets, threshold):
        c = sim, i0, i1
        candidates[0][i0].append(c)
        candidates[1][i1].append(c)
            
    # Take the best k candidates for each record and count them
    pair_counter: Counter[_Similarity] = collections.Counter()
    for dset_cands in candidates:
        for unfiltered_cands in dset_cands.values():
            best = heapq.nsmallest(k, unfiltered_cands,
                                   key=lambda x:(-x[0], *x[1:]))
            pair_counter.update(best)

    sims: FloatArrayType = array('d')
    indices0: IntArrayType = array('I')
    indices1: IntArrayType = array('I')
    for pair in pair_counter:
        if pair_counter[pair] == 2:
            sim, i0, i1 = pair
            sims.append(sim)
            indices0.append(i0)
            indices1.append(i1)

    return sims, (indices0, indices1)


def _smc_similarity_no_k(
    datasets: Sequence[Dataset[bitarray]],
    threshold: float
) -> Tuple[FloatArrayType, Tuple[IntArrayType, ...]]:
    sims: FloatArrayType = array('d')
    indices0: IntArrayType = array('I')
    indices1: IntArrayType = array('I')
    
    for sim, i0, i1 in sorted(
            _smc_sims_gt_threshold(datasets, threshold),
            key=lambda x: (-x[0], *x[1:])):
        sims.append(sim)
        indices0.append(i0)
        indices1.append(i1)

    return sims, (indices0, indices1)


def simple_matching_coefficient(
    datasets: Sequence[Dataset[bitarray]],
    threshold: float,
    k: Optional[int]
) -> Tuple[FloatArrayType, Tuple[IntArrayType, ...]]:
    """Find Simple Matching Coefficients (SMCs) of CLKs.

    The SMC of two binary strings is defined as
    1 - (Hamming distance) / (string length).

    :param datasets: A length 2 sequence of datasets. A dataset is a
        sequence of bitarrays.
    :param threshold: Pairs whose similarity is above this value may be
        a match.
    :param k: We only return the top k candidates for every record. Set
        to None to return all candidates.
    
    :return: A 2-tuple of similarity scores and indices. The similarity
        scores are an array of floating-point values. The indices are a
        2-tuple of arrays of integers.
    """
    if len(datasets) != 2:
        msg = 'only binary matching is currently supported'
        raise NotImplementedError(msg)

    if k is None:
        sims, indices = _smc_similarity_no_k(datasets, threshold)
    else:
        sims, indices = _smc_similarity_k(datasets, threshold, k)

    # Quick sanity checks
    assert len(indices) == 2
    assert len(sims) == len(indices[0]) == len(indices[1])

    return sims, indices


@_deprecated_decorator(replacement='similarities.simple_matching_coefficient')
def hamming_similarity(
    datasets: Sequence[Dataset[bitarray]],
    threshold: float,
    k: Optional[int]
) -> Tuple[FloatArrayType, Tuple[IntArrayType, ...]]:
    return simple_matching_coefficient(datasets, threshold, k)
hamming_similarity.__doc__ = simple_matching_coefficient.__doc__



"""Blocking functions.

These functions are a probabilistic optimisation of similarity
computation. They aim to avoid comparing all records to each other by
only comparing records that are more likely to be a match. Blocking
generally reduces recall, but does not reduce precision.

Records are split into discrete buckets. A record may belong to multiple
buckets. Generally, two records are compared by a similarity function if
they share at least one bucket.
"""

import itertools as _itertools
import random as _random
import typing as _typing

import bitarray as _bitarray

import anonlink.typechecking as _typechecking


_T = _typing.TypeVar('_T')
def _evalf(__funcs : _typing.Iterable[_typing.Callable[..., _T]],
           *args,
           **kwargs) -> _typing.Iterable[_T]:
    """Apply a number of functions to the same arguments.

    Think of this as reverse map. Instead of having one function and
    multiple data, you have multiple functions and one piece of data.
    """
    return (f(*args, **kwargs) for f in __funcs)


# https://mypy.readthedocs.io/en/stable/more_types.html#function-overloading
@_typing.overload
def block_and(
    __funcs: _typing.Iterable[
        _typechecking.BlockingFunction[_typechecking.Record]]
) -> _typechecking.BlockingFunction[_typechecking.Record]:
    ...
@_typing.overload
def block_and(
    __func1: _typechecking.BlockingFunction[_typechecking.Record],
    __func2: _typechecking.BlockingFunction[_typechecking.Record],
    *funcs: _typechecking.BlockingFunction[_typechecking.Record]
) -> _typechecking.BlockingFunction[_typechecking.Record]:
    ...
def block_and(*args) -> _typechecking.BlockingFunction[_typechecking.Record]:
    """Return conjunction of multiple blocking functions.

    Records share a block if they share a block in all of the functions.

    :param funcs: Functions whose conjunction we return.

    :return: The blocking function.
    """
    if len(args) == 0:
        raise TypeError('expected at least 1 argument, got 0')
    elif len(args) == 1:
        funcs = tuple(args[0])
    else:
        funcs = args

    def block_and_inner(
        dataset_index: int,
        record_index: int,
        hash_: _typechecking.Record
    ) -> _typing.Iterable[_typing.Hashable]:
        funcs_eval = _evalf(funcs, dataset_index, record_index, hash_)
        return _itertools.product(*funcs_eval)

    return block_and_inner


@_typing.overload
def block_or(
    __funcs: _typing.Iterable[
        _typechecking.BlockingFunction[_typechecking.Record]]
) -> _typechecking.BlockingFunction[_typechecking.Record]:
    ...
@_typing.overload
def block_or(
    __func1: _typechecking.BlockingFunction[_typechecking.Record],
    __func2: _typechecking.BlockingFunction[_typechecking.Record],
    *funcs: _typechecking.BlockingFunction[_typechecking.Record]
) -> _typechecking.BlockingFunction[_typechecking.Record]:
    ...
def block_or(*args) -> _typechecking.BlockingFunction[_typechecking.Record]:
    """Return disjunction of multiple blocking functions.

    Records share a block if they share a block in any of the functions.

    :param funcs: Functions whose disjunction we return.

    :return: The blocking function.
    """
    if len(args) == 0:
        raise TypeError('expected at least 1 argument, got 0')
    elif len(args) == 1:
        funcs = tuple(args[0])
    else:
        funcs = args

    def block_or_inner(
        dataset_index: int,
        record_index: int,
        hash_: _typechecking.Record
    ) -> _typing.Iterable[_typing.Hashable]:
        funcs_eval = _evalf(funcs, dataset_index, record_index, hash_)
        funcs_enum = (
            zip(_itertools.repeat(i), f) for i, f in enumerate(funcs_eval))
        return _itertools.chain.from_iterable(funcs_enum)

    return block_or_inner


def bit_blocking(
    g: int,
    r: int,
    *,
    seed: _typing.Any = None
) -> _typechecking.BlockingFunction[_bitarray.bitarray]:
    """Block on bits of the hash.

    This blocking type has the advantage that it doesn't require any
    auxilliary information. Despite this, it is still probabilistic and
    may decrease recall.

    :param int r: The number of bits to use per block.
    :param int g: The number of blocks per record.
    :param seed: Optional seed for pseudorandom number generation. Used
        to keep the results consistent.

    :return: The blocking function.
    """
    if g < 1:
        msg = f'g is expected to be positive but is {g}'
        raise ValueError(msg)
    if r < 1:
        msg = f'r is expected to be positive but is {r}'
        raise ValueError(msg)

    hash_len: _typing.Optional[int] = None
    hash_indices: _typing.Optional[_typing.Sequence[_typing.Sequence[int]]] \
        = None

    def bit_blocking_inner(
        dataset_index: int,
        record_index: int,
        hash_: _bitarray.bitarray
    ) -> _typing.Iterable[_typing.Hashable]:
        nonlocal hash_len, hash_indices

        this_hash_len = len(hash_)
        if hash_indices is None:
            # First call.
            hash_len = this_hash_len

            rng = _random.Random(seed)
            # Future: consider keeping track of already chosen indices
            # and avoiding them. This may improve performance by making
            # tables more independent.
            hash_indices = tuple(rng.sample(range(hash_len), r)
                                 for _ in range(g))

        assert hash_len is not None
        if this_hash_len != hash_len:
            raise ValueError('inconsistent hash length')

        assert hash_indices is not None
        for i, table_indices in enumerate(hash_indices):
            vals = map(hash_.__getitem__, table_indices)

            # We need to turn this iterable of bools into something
            # hashable. An int will do just fine.
            table_block = sum(b << j for j, b in enumerate(vals))

            # Distinguish between tables.
            yield table_block * len(hash_indices) + i

    return bit_blocking_inner


def continuous_blocking(
    radius: float,
    source: _typing.Sequence[_typing.Sequence[float]]
) -> _typechecking.BlockingFunction[_typechecking.Record]:
    """Block on continuous variables.

    Split the real number line into overlapping blocks. A quantity is in
    exactly two blocks. Two quantities within `radius` of each other are
    guaranteed to have at least one block in common.

    :param float radius: Two quantities within this radius are
        guaranteed to have at least one block in common.
    :param source: The source as a sequence of sequences of reals.

    :return: The blocking function.
    """
    if radius <= 0:
        raise ValueError(f'radius should be positive, got {radius}')

    def continuous_blocking_inner(
        dataset_index: int,
        record_index: int,
        hash_: _typechecking.Record
    ) -> _typing.Iterable[_typing.Hashable]:
        x = source[dataset_index][record_index]
        r = radius

        bucket_1 = int(x // (2 * r))
        bucket_2 = int((x + r) // (2 * r))

        # Ensure bucket_1 is odd and bucket_2 is even to distinguish.
        return (bucket_1 * 2, bucket_2 * 2 + 1)

    return continuous_blocking_inner


def list_blocking(
    source: _typing.Sequence[_typing.Sequence[_typing.Hashable]]
) -> _typechecking.BlockingFunction[_typechecking.Record]:
    """Block on the contents of a sequence of sequences.

    This is a simple convenience function. Every inner sequence
    corresponds to a dataset. Every element of an inner sequence
    corresponds to a record.

    :param source: Sequence of sequence of blocks.

    :return: The blocking function.
    """
    def list_blocking_inner(
        dataset_index: int,
        record_index: int,
        hash_: _typechecking.Record
    ) -> _typing.Iterable[_typing.Hashable]:
        return (source[dataset_index][record_index],)

    return list_blocking_inner



import pkg_resources

from anonlink import blocking
from anonlink import candidate_generation
from anonlink import concurrency
from anonlink import serialization
from anonlink import similarities
from anonlink import solving
from anonlink import stats
from anonlink import typechecking

__version__ = pkg_resources.get_distribution('anonlink').version
__author__ = 'Data61'



import platform

from setuptools import setup, Extension, find_packages
import os

try:
    from Cython.Build import cythonize, build_ext
except ImportError:
    # No Cython but we may have pre-converted the files
    maybe_cythonize = lambda x: x
    cython_cpp_ext = 'cpp'
else:
    maybe_cythonize = cythonize
    cython_cpp_ext = 'pyx'

here = os.path.abspath(os.path.dirname(__file__))

requirements = [
        "bitarray>=2.5",
        "numpy>=1.19",
        "mypy-extensions>=0.4",
        "Cython>=0.29"
    ]

test_requirements = [
        "pytest",
        "pytest-timeout",
        "pytest-cov",
        "codecov",
        "hypothesis",
        "clkhash>=0.16.1",
    ]

current_os = platform.system()
if current_os == "Windows":
    extra_compile_args = ['/std:c++17', '/O2']
    extra_link_args = []
else:
    extra_compile_args = ['-O3', '-std=c++11']
    extra_link_args = []

extensions = [
    Extension(
        name="solving._multiparty_solving",
        sources=["anonlink/solving/_multiparty_solving." + cython_cpp_ext,
                 "anonlink/solving/_multiparty_solving_inner.cpp"],
        include_dirs=["anonlink/solving"],
        language="c++",
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_compile_args,
        define_macros=[('NDEBUG', None)]
        ),
    Extension(
        name="similarities._dice",
        sources=[
            "anonlink/similarities/_dice." + cython_cpp_ext
        ],
        include_dirs=["anonlink/similarities"],
        language="c++",
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        define_macros=[('NDEBUG', None)]
        )
]

with open('README.rst', 'r', encoding='utf-8') as f:
    readme = f.read()

setup(
    name="anonlink",
    version='0.15.1',
    description='Anonymous linkage using cryptographic hashes and bloom filters',
    long_description=readme,
    long_description_content_type='text/x-rst',
    url='https://github.com/data61/anonlink',
    license='Apache',
    setup_requires=["pytest-runner"],
    install_requires=requirements,
    tests_require=test_requirements,
    extras_require={
        "test": test_requirements
    },
    packages=find_packages(exclude=[
        'tests'
    ]),
    package_data={'anonlink': []},
    ext_modules=maybe_cythonize(extensions),
    ext_package="anonlink",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Security :: Cryptography",
    ],
    zip_safe=False,
)



import functools as _functools
import typing as _typing
import warnings as _warnings

import mypy_extensions as _mypy_extensions


def _warn_deprecated(
    module_name: str,
    f_name: str,
    replacement: _typing.Optional[str]
) -> None:
    msg = f'anonlink.{module_name}.{f_name} has been deprecated '
    if replacement is not None:
        msg += f'(use anonlink.{replacement} instead)'
    else:
        msg += 'without replacement'
    _warnings.warn(msg, DeprecationWarning, stacklevel=3)


_WrappedF = _typing.TypeVar('_WrappedF',
                            bound=_typing.Callable[..., _typing.Any])
def make_decorator(
    module_name: str
) -> _typing.Callable[..., _typing.Any]:
    @_typing.overload
    def deprecate_decorator(
        f: None = None,
        *,
        replacement: _typing.Optional[str] = None
    ) -> _typing.Callable[[_WrappedF], _WrappedF]:
        ...
    @_typing.overload
    def deprecate_decorator(
        f: _WrappedF,
        *,
        replacement: _typing.Optional[str] = None
    ) -> _WrappedF:
        ...
    def deprecate_decorator(
        f: _typing.Optional[_WrappedF] = None,
        *,
        replacement: _typing.Optional[str] = None
    ) -> _typing.Union[_WrappedF, _typing.Callable[[_WrappedF], _WrappedF]]:
        def deprecate_decorator_inner(f: _WrappedF) -> _WrappedF:
            @_functools.wraps(f)
            def f_inner(*args, **kwargs):
                _warn_deprecated(module_name, f.__name__, replacement)
                return f(*args, **kwargs)
            return _typing.cast(_WrappedF, f_inner)
        # Use decorator with or without arguments.
        if f is None:
            return deprecate_decorator_inner
        else:
            return deprecate_decorator_inner(f)
    return deprecate_decorator



"""Serialisation of candidate pairs.

The file format is not fixed and is subject to change. Backwards
compatibility will be attempted, but not guaranteed.
"""

import array as _array
import heapq as _heapq
import io as _io
import itertools as _itertools
import operator as _operator
import struct as _struct
import typing as _typing

import anonlink.typechecking as _typechecking

# FILE FORMAT
#   This is subject to change.
#   The format is composed of a header and a sequence of entries. The
# header specifies the version of the format and the sizes of floating-
# -point and integer types used in the entries. Every entry is composed
# of a similarity score, two dataset indices, and two record indices.
#   The header is composed of a version (1 byte) and the format of the
# candidate pairs. This format consists of the number of bytes in a
# similarity (1 byte), the number of bytes in a dataset index (1 byte),
# and the number in a record index (1 byte). In total, the header is
# composed of 4 bytes.
#   An entry is composed of a similarity as a IEEE floating-point value,
# two dataset indices, and two record indices. Their sizes are
# determined by the header. For example, if the similarity is an IEEE
# double-precision floating-point value, the size noted by the header
# will be 8 [bytes].
#   All integers are unsigned. All values are *little-endian* since
# anonlink is realistically only ever run on little-endian machines.
# (This code will still run on big-endian machines, implicitly
# performing byteorder conversion.) The number of candidate matchings is
# not included on purpose: if needed, it can be computed from the length
# of the file.
#   Similarity scores are stored as a structure of arrays when in
# memory, but we serialise them as an array of structures. This makes it
# much easier to merge two serialised files.

# https://docs.python.org/3/library/struct.html#format-characters
_STRUCT_UINT_LEN_TO_FMT = {1: 'B', 2: 'H', 4: 'L', 8: 'Q'}
_STRUCT_FLOAT_LEN_TO_FMT = {2: 'e', 4: 'f', 8: 'd'}

# https://docs.python.org/3/library/array.html#module-array
_ARRAY_UINT_TYPES = 'BIHLQ'
_ARRAY_FLOAT_TYPES = 'df'
_ARRAY_UINT_LEN_TO_FMT = {_array.array(t).itemsize: t
                          for t in _ARRAY_UINT_TYPES}
_ARRAY_FLOAT_LEN_TO_FMT = {_array.array(t).itemsize: t
                           for t in _ARRAY_FLOAT_TYPES}

_HEADER_STRUCT = _struct.Struct('<BBBB')

_CandidatePair = _typing.Tuple[float, int, int, int, int]
_CandidatePairIter = _typing.Iterable[_CandidatePair]


def _entry_struct(
    sim_t_size: int,
    dset_i_t_size: int,
    rec_i_t_size: int
) -> _struct.Struct:
    try:
        sim_t = _STRUCT_FLOAT_LEN_TO_FMT[sim_t_size]
    except KeyError:
        msg = f'floats of {sim_t_size} bytes are not supported'
        raise ValueError(msg) from None
    try:
        dset_i_t = _STRUCT_UINT_LEN_TO_FMT[dset_i_t_size]
    except KeyError:
        msg = f'indices of {dset_i_t_size} bytes are not supported'
        raise ValueError(msg) from None
    try:
        rec_i_t = _STRUCT_UINT_LEN_TO_FMT[rec_i_t_size]
    except KeyError:
        msg = f'indices of {rec_i_t_size} bytes are not supported'
        raise ValueError(msg) from None
    # Every entry has: a similarity score, two datset indices, and two
    # record indices.
    return _struct.Struct(f"<{sim_t}2{dset_i_t}2{rec_i_t}")
    

def _bytes_iter_from_iterable(
    sim_t_size: int,
    dset_i_t_size: int,
    rec_i_t_size: int,
    entry_struct: _struct.Struct,
    iterable: _CandidatePairIter
) -> _typing.Iterable[bytes]:
    return _itertools.chain(
        (_HEADER_STRUCT.pack(1, sim_t_size, dset_i_t_size, rec_i_t_size),),
        _itertools.starmap(entry_struct.pack, iterable))


def _write_bytes_iter(
    f: _typing.BinaryIO,
    bytes_iter: _typing.Iterable[bytes]
) -> int:
    # Relies on the side effects of `f.write`. `f.write` returns the
    # number of bytes written. `sum` exhausts the iterator.
    return sum(map(f.write, bytes_iter))


def _entries_iterable(
    f: _typing.BinaryIO,
    entry_struct: _struct.Struct
) -> _CandidatePairIter:
    buffers = map(f.read, _itertools.repeat(entry_struct.size))
    nonempty_buffers = _itertools.takewhile(bool, buffers)
    try:
        yield from _typing.cast(_CandidatePairIter, map(entry_struct.unpack, nonempty_buffers))
    except _struct.error:
        raise ValueError('ran out of input') from None


def _make_buffered(
    f: _typing.BinaryIO
) -> _typing.BinaryIO:
    # Make sure that f.read(n) returns exactly n bytes.
    if isinstance(f, _io.RawIOBase):
        f = _io.BufferedReader(f)
    # If f is neither BufferedReader nor RawIOBase then we shrug and see
    # what happens...
    return f


def _load_header_and_check_version(
    f: _typing.BinaryIO  # Must be buffered
) -> _typing.Tuple[int, int, int]:
    header_bytes = f.read(_HEADER_STRUCT.size)
    if len(header_bytes) != _HEADER_STRUCT.size:
        raise ValueError('ran out of input')
    version, sim_t_size, dset_i_t_size, rec_i_t_size = (
        _HEADER_STRUCT.unpack(header_bytes))
    if version != 1:
        raise ValueError('unsupported version of serialized file')
    return sim_t_size, dset_i_t_size, rec_i_t_size


def _load_to_iter_with_sizes(
    f: _typing.BinaryIO
) -> _typing.Tuple[_CandidatePairIter, int, int, int, int]:
    f = _make_buffered(f)
    sim_t_size, dset_i_t_size, rec_i_t_size = _load_header_and_check_version(f)

    # This may throw ValueError if the specified format isn't supported
    # on this platform.
    entry_struct = _entry_struct(sim_t_size, dset_i_t_size, rec_i_t_size)

    return (_entries_iterable(f, entry_struct),
            sim_t_size, dset_i_t_size, rec_i_t_size,
            entry_struct.size)


def load_to_iterable(
    f: _typing.BinaryIO
) -> _CandidatePairIter:
    """Load candidate pairs from file as an iterable.

    This function does not load all the candidate pairs into memory at
    once.

    :param f: Binary stream to read from.

    :return: An iterable of 5-tuples. Each 5-tuple represents one
        candidate pair and is composed of the similarity, the index of
        the first record's dataset, the index of the second record's
        dataset, the index of the first record within its dataset, and
        the index of the second record within its dataset.
    """
    iterable, _, _, _, _ = _load_to_iter_with_sizes(f)
    return iterable


def _file_size(entry_struct, entries):
    return _HEADER_STRUCT.size + entry_struct.size * entries


def dump_candidate_pairs_iter(
    candidate_pairs: _typechecking.CandidatePairs
) -> _typing.Tuple[_typing.Iterable[bytes], int]:
    """Dump candidate pairs as an iterable of bytes objects.

    No guarantees are made about the size of those bytes objects.

    :param candidate_pairs: Candidate pairs, as returned by a similarity
        function or `load_candidate_pairs`.

    :return: 2-tuple containing an iterable of bytes objects and the
        length of the dump as an integer.
    """
    sims, (dset_is0, dset_is1), (rec_is0, rec_is1) = candidate_pairs
    entries = len(sims)
    assert (entries
            == len(dset_is0) == len(dset_is1)
            == len(rec_is0) == len(rec_is1))

    sim_t_size = sims.itemsize
    dset_i_t_size = max(dset_is0.itemsize, dset_is1.itemsize)
    rec_i_t_size = max(rec_is0.itemsize, rec_is1.itemsize)
    entry_struct = _entry_struct(sim_t_size, dset_i_t_size, rec_i_t_size)

    iterable = zip(sims, dset_is0, dset_is1, rec_is0, rec_is1)
    bytes_iter = _bytes_iter_from_iterable(
        sim_t_size, dset_i_t_size, rec_i_t_size,
        entry_struct,
        iterable)

    file_size = _file_size(entry_struct, entries)

    return bytes_iter, file_size


def dump_candidate_pairs(
    candidate_pairs: _typechecking.CandidatePairs,
    f: _typing.BinaryIO
) -> int:
    """Dump candidate pairs to file.

    :param f: Binary stream to write to.
    :param candidate_pairs: Candidate pairs, as returned by a similarity
        function or `load_candidate_pairs`.

    :return: Number of bytes written.
    """
    bytes_iter, _ = dump_candidate_pairs_iter(candidate_pairs)
    return _write_bytes_iter(f, bytes_iter)


def load_candidate_pairs(f: _typing.BinaryIO) -> _typechecking.CandidatePairs:
    """Load candidate pairs from file.

    :param f: Binary stream to read from.

    :return: Candidate pairs, compatible with the type returned from a
        similarity function.
    """
    iterable_with_sizes = _load_to_iter_with_sizes(f)
    iterable, sim_t_size, dset_i_t_size, rec_i_t_size, _ = iterable_with_sizes

    try:
        sim_t = _ARRAY_FLOAT_LEN_TO_FMT[sim_t_size]
    except KeyError:
        msg = f'floats of {sim_t_size} bytes are not supported'
        raise ValueError(msg) from None
    try:
        dset_i_t = _ARRAY_UINT_LEN_TO_FMT[dset_i_t_size]
    except KeyError:
        msg = f'indices of {dset_i_t_size} bytes are not supported'
        raise ValueError(msg) from None
    try:
        rec_i_t = _ARRAY_UINT_LEN_TO_FMT[rec_i_t_size]
    except KeyError:
        msg = f'indices of {rec_i_t_size} bytes are not supported'
        raise ValueError(msg) from None

    # Future: preallocate memory according to length of file. This is
    # not possible in pure Python, but can be done in Cython.
    sims: _typechecking.FloatArrayType = _array.array(sim_t)
    dset_is0: _typechecking.IntArrayType = _array.array(dset_i_t)
    dset_is1: _typechecking.IntArrayType = _array.array(dset_i_t)
    rec_is0: _typechecking.IntArrayType = _array.array(rec_i_t)
    rec_is1: _typechecking.IntArrayType = _array.array(rec_i_t)
    arrays = sims, dset_is0, dset_is1, rec_is0, rec_is1

    # Rely on side side-effecting function passed to append.
    # any exhausts the iterator since array.append returns None.
    any(map(_array.array.append,  # type: ignore  # Just give up, Mypy.
            _itertools.cycle(arrays),
            _itertools.chain.from_iterable(iterable)))

    return sims, (dset_is0, dset_is1), (rec_is0, rec_is1)


def _number_entries(file_size, entry_size):
    entries, remainder = divmod(file_size - _HEADER_STRUCT.size, entry_size)
    if remainder:
        raise ValueError('invalid file: number of entries is non-integer')
    return entries 


def merge_streams_iter(
    files_in: _typing.Iterable[_typing.BinaryIO],
    *,
    sizes: _typing.Optional[_typing.Iterable[int]] = None
) -> _typing.Tuple[_typing.Iterable[bytes], int]:
    """Merge multiple files with candidate pairs to iterable of bytes.

    This function preserves the candidate pairs' sorted order. It avoids
    loading everything into memory.

    The field sizes in the resulting file will be chosen so no
    information is lost.
     
    Note that you cannot simply concatenate two files to obtain a valid
    candidate pairs dump.

    :param files_in: Sequence of files to read from.
    :param sizes: Optional iterable of file sizes. Permits us to compute
        the number of bytes in the returned iterator.

    :return: 2-tuple containing an iterable of bytes objects and 
        (optionally if the `sizes` parameter was provided) the length of
        the merged file as an integer.
    """
    if not files_in:
        raise ValueError('no files provided')

    iterables_with_sizes = map(_load_to_iter_with_sizes, files_in)
    file_iterables, *field_sizes = zip(*iterables_with_sizes)
    file_sim_t_size, file_dset_i_t_size, file_rec_i_t_size, file_entry_size \
        = field_sizes

    sim_t_size = max(file_sim_t_size)
    dset_i_t_size = max(file_dset_i_t_size)
    rec_i_t_size = max(file_rec_i_t_size)
    
    entry_struct = _entry_struct(sim_t_size, dset_i_t_size, rec_i_t_size)

    # Sort in decreasing order of similarities. Tiebreak with dataset
    # indices and then with record indices, in increasing order.
    sorted_iterable = _heapq.merge(*file_iterables,
                                   key=lambda x: (-x[0],) + x[1:])
    bytes_iter = _bytes_iter_from_iterable(
        sim_t_size, dset_i_t_size, rec_i_t_size,
        entry_struct,
        sorted_iterable)

    if sizes is not None:
        entries_num = sum(map(_number_entries, sizes, file_entry_size))
        file_size = _file_size(entry_struct, entries_num)
    else:
        file_size = None

    return bytes_iter, file_size


def merge_streams(
    files_in: _typing.Iterable[_typing.BinaryIO],
    f_out: _typing.BinaryIO
) -> int:
    """Merge multiple files with serialised candidate pairs.

    This function preserves the candidate pairs' sorted order. It avoids
    loading everything into memory.

    The field sizes in the resulting file will be chosen so no
    information is lost.
     
    Note that you cannot simply concatenate two files to obtain a valid
    candidate pairs dump.

    :param files_in: Sequence of files to read from.
    :param f_out: Binary stream write the merged candidate pairs to.

    :return: Number of bytes written.
    """
    bytes_iter, _ = merge_streams_iter(files_in)
    return _write_bytes_iter(f_out, bytes_iter)



from array import array
from itertools import chain, groupby, repeat
from typing import Optional, Sequence, Tuple

from bitarray import bitarray

from anonlink.similarities import _dice
from anonlink.similarities._utils import (sort_similarities_inplace,
                                          to_bitarrays)
from anonlink.typechecking import FloatArrayType, IntArrayType

__all__ = ['dice_coefficient_accelerated']


def _all_equal(iterable):
    # https://docs.python.org/3/library/itertools.html#itertools-recipes
    """Return True if all the elements are equal to each other."""
    g = groupby(iterable)
    return next(g, True) and not next(g, False)


def dice_coefficient_accelerated(
    datasets: Sequence[Sequence[bitarray]],
    threshold: float,
    k: Optional[int] = None
) -> Tuple[FloatArrayType, Tuple[IntArrayType, ...]]:
    """Find Dice coefficients of CLKs.

    This version uses a CPP implementation which will take advantage of
    native x86 popcount instructions.

    We assume all filters are the same length and that this length is a
    multiple of 8 bits.

    The returned pairs are sorted in decreasing order of similarity,
    then in increasing order of the record index in the first dataset,
    and then in increasing order of the record index in the second
    dataset.

    :param datasets: A length 2 sequence of datasets. A dataset is a
        sequence of bitarrays.
    :param threshold: The similarity threshold. We accept pairs that
        have similarity of at least this value.
    :param k: Only permit this many candidate pairs per dataset pair
        per record. Set to `None` to permit all pairs above with
        similarity at least `threshold`.

    :raises NotImplementedError: If an unsupported length filter is
        provided.

    :raises ValueError: If different filter lengths are provided.

    :return: A 2-tuple of similarity scores and indices. The similarity
        scores are an array of floating-point values. The indices are a
        2-tuple of arrays of integers.
    """
    n_datasets = len(datasets)
    if n_datasets < 2:
        raise ValueError(f'not enough datasets (expected 2, got {n_datasets})')
    elif n_datasets > 2:
        raise NotImplementedError(
            f'too many datasets (expected 2, got {n_datasets})')
    filters0, filters1 = datasets
    filters0 = to_bitarrays(filters0)
    filters1 = to_bitarrays(filters1)

    # Create empty output arrays
    result_sims: FloatArrayType = array('d')
    result_indices0: IntArrayType = array('I')
    result_indices1: IntArrayType = array('I')

    if not filters0 or not filters1:
        # Empty result of the correct type.
        return result_sims, (result_indices0, result_indices1)

    length_f0, length_f1 = map(len, datasets)

    # There's no sense in having k > length_f1. Also, k is used to
    # allocate memory below, so we need to protect against an
    # out-of-memory DoS if k is of untrustworthy origin.
    if k is None or k > length_f1:
        k = length_f1

    if not _all_equal(map(len, chain(filters0, filters1))):
        raise ValueError('inconsistent filter length')
    filter_bits = len(filters0[0])
    if filter_bits % 8:
        msg = (f'only filters whose length in bits is a multiple of 8 '
               f'are currently supported (got filter with length '
               f'{filter_bits})')
        raise NotImplementedError(msg)
    filter_bytes = filter_bits // 8
    # Python char arrays of all filters from filters0 and filter1
    carr0, carr1 = array('b'), array('b')

    carr0.frombytes(b''.join(memoryview(f) for f in filters0))
    carr1.frombytes(b''.join(memoryview(f) for f in filters1))

    # Only worth popcounting in C for a large number of filters.
    # Current threshold was found by trying out different values while benchmarking
    POPCOUNT_NATIVE_THRESHOLD = 10000
    if len(filters1) < POPCOUNT_NATIVE_THRESHOLD:
        c_popcounts = array('I', [f.count() for f in filters1])
    else:
        c_popcounts = _dice.popcount_arrays(carr1, filter_bytes)

    _dice.dice_many_to_many(
        carr0, carr1, length_f0, length_f1, c_popcounts, filter_bytes, k,
        threshold, result_sims, result_indices0, result_indices1)

    sort_similarities_inplace(result_sims, result_indices0, result_indices1)

    return result_sims, (result_indices0, result_indices1)




import collections as _collections
import itertools as _itertools
import typing as _typing

import anonlink.typechecking as _typechecking


def probabilistic_greedy_solve_python(
    candidates: _typechecking.CandidatePairs,
    *,
    merge_threshold: float = .5,
    deduplicated: bool = True
) -> _typechecking.MatchGroups:
    """Select matches using the probabilistic greedy algorithm.

    This solver is similar to `greedy_solve_native`, but may match
    records that do not have corresponding candidate pairs when there is
    enough other evidence that they are a match. This usually results in
    better accuracy.

    We assign each record to exactly one 'group' of records that are
    pairwise matched together. We iterate over `candidates` in order of
    decreasing similarity. When we encounter a pair of records (let s be
    their similarity) that do not already belong to the same group, we
    merge their groups iff (a) the proportion of pairs that are
    permitted to be matched in the groups' records is at least
    `merge_threshold`, and (b) the proportion of pairs whose similarity
    is above s in the groups' records is at least `merge_threshold`. The
    latter requirement is only for tie-breaking: we prefer groups whose
    minimum pairwise similarity is highest. Any group merge rejected due
    to requirement (b) will be revisited later with that requirement
    relaxed.

    :param tuple candidates: Candidates, as returned by
        `find_candidates`.
    :param float merge_threshold: Proportion of agreement between groups
        required to merge them. In (0, 1].
            When considering whether to merge two groups, we compute the
        groups' Cartesian product. We count the pairs in that product
        whose similarity is above the threshold. We merge the groups if
        our count is at least `merge_threshold` of the total number of
        pairs between the two groups.
            For example, suppose that we have two groups of records: 
        {A, B} and {C, D}. Let A-C, A-D and B-C be candidate pairs.
        Setting `merge_threshold` to any value not greater than .75 will
        permit the two groups to be merged since they share three
        candidate pairs out of all four pairs. Even though the
        similarity of B-D is below the threshold, there is enough
        evidence in the other pairs to conclude that it must be a match.
        The `merge_threshold` can be thought of as the inverse of the
        weight we place on that evidence.
            Higher values increase precision, but decrease recall.
            A `merge_threshold` lower than one permits two records to be
        matched without having a corresponding candidate pair. Set it to
        1.0 to ensure that two records will only be matched if their
        similarity is above the similarity threshold.
    :param bool deduplicated: When True, two records that belong to the
        same dataset will never be in the same group. Default True.

    :return: An sequence of groups. Each group is an sequence of
        records. Two records are in the same group iff they represent
        the same entity. Here, a record is a two-tuple of dataset index
        and record index.
    """
    if merge_threshold < 0 or merge_threshold > 1:
        raise ValueError(
            f'merge_threshold must be between 0 and 1 (got {merge_threshold})') 

    sims, dset_is, rec_is = candidates
    if len(dset_is) != len(rec_is):
        raise ValueError('inconsistent shape of index arrays')
    if len(dset_is) != 2:
        raise NotImplementedError('only binary solving is supported')

    dset_is0, dset_is1 = dset_is
    rec_is0, rec_is1 = rec_is
    if not (len(sims)
            == len(dset_is0) == len(dset_is1)
            == len(rec_is0) == len(rec_is1)):
        raise ValueError('inconsistent shape of index arrays')

    # Map (dataset index, record index) to its group.
    matches: _typing.Dict[_typing.Tuple[int, int],
                          _typing.List[_typing.Tuple[int, int]]] = {}

    # Each group is a set of records. We merge two groups when 
    # pair of their records is matchable. A pair is matchable if we have
    # seen it. Store the number of matchable pairs between two groups.
    # This is a sparse matrix: the default number of matchable pairs is
    # 0. Since the groups themselves are not hashable, we use their id
    # as the key.
    matchable_pairs: _typing.DefaultDict[int, _typing.Counter[int]] \
        = _collections.defaultdict(_collections.Counter)

    for dset_i0, dset_i1, rec_i0, rec_i1 in zip(
            dset_is0, dset_is1, rec_is0, rec_is1):
        i0 = dset_i0, rec_i0
        i1 = dset_i1, rec_i1

        if i0 == i1:
            continue

        if i0 in matches and i1 in matches:
            # Both records are assigned to a group.
            i0_matches = matches[i0]
            i1_matches = matches[i1]
            i0_mid = id(i0_matches)
            i1_mid = id(i1_matches)

            if i0_mid == i1_mid:
                continue

            # Check if mergeable. matchable_pairs[i0_mid][i1_mid] is the
            # number of pairs they have in common not including the
            # current pair--we add one to include it. The total number
            # of pairs is len(i0_matches) * len(i1_matches)). When this
            # is the number of matchable pairs, then every pair is
            # matchable.
            overlap = matchable_pairs[i0_mid][i1_mid] + 1
            total_pairs = len(i0_matches) * len(i1_matches)
            duplicates_ok = (not deduplicated or all(m0 != m1
                                                     for m0, _ in i0_matches
                                                     for m1, _ in i1_matches))
            if overlap >= merge_threshold * total_pairs and duplicates_ok:
                # Optimise by always extending the bigger group.
                if len(i0_matches) < len(i1_matches):
                    i0, i1 = i1, i0
                    i0_mid, i1_mid = i1_mid, i0_mid
                    i0_matches, i1_matches = i1_matches, i0_matches
                # Merge groups.
                i0_matches.extend(i1_matches)
                matches.update(zip(i1_matches, _itertools.repeat(i0_matches)))
                
                # Update matchable pairs.
                del matchable_pairs[i0_mid][i1_mid]
                del matchable_pairs[i1_mid][i0_mid]
                for j_mid, j_count in matchable_pairs[i1_mid].items():
                    matchable_pairs[i0_mid][j_mid] += j_count
                    matchable_pairs[j_mid][i0_mid] += j_count
                    del matchable_pairs[j_mid][i1_mid]
                del matchable_pairs[i1_mid]
                if not matchable_pairs[i0_mid]:  # Empty. Can delete.
                    del matchable_pairs[i0_mid]

            else:
                # Don't merge. Mark: they have another edge in common.
                matchable_pairs[i0_mid][i1_mid] += 1
                matchable_pairs[i1_mid][i0_mid] += 1
            continue

        if i0 not in matches and i1 in matches:
            i0, i1 = i1, i0
            # Symmetry. Fall through.
        if i0 in matches and i1 not in matches:
            # i0 is in a group, but i1 is not.
            # See if we may assign i0 to that group.
            i0_matches = matches[i0]
            overlap = 1
            total_pairs = len(i0_matches)
            duplicates_ok = not deduplicated or all(m0 != i1[0]
                                                    for m0, _ in i0_matches)
            if overlap >= merge_threshold * total_pairs and duplicates_ok:
                # i0 is a group of 1, so trivially we can merge.
                i0_matches.append(i1)
                matches[i1] = i0_matches
            else:
                # i0 is a group of >1. i1 is not in a group, so this is
                # the first time we're seeing it. Hence, it is not
                # matchable with the other elements of i0.
                i1_matches = [i1]
                matches[i1] = i1_matches

                matchable_pairs[id(i1_matches)][id(i0_matches)] = 1
                matchable_pairs[id(i0_matches)][id(i1_matches)] = 1
            continue

        if i0 not in matches and i1 not in matches:
            duplicates_ok = not deduplicated or i0[0] != i1[0]
            if duplicates_ok:
                # Neither is in a group, so let's just make one.
                matches[i0] = matches[i1] = [i0, i1]
            continue

        raise RuntimeError('non-exhaustive cases')
                
    # Return all nontrivial groups without duplication
    deduplicated_groups = {id(group): group
                           for group in matches.values()
                           if len(group) > 1}
    return tuple(map(tuple, deduplicated_groups.values()))


def greedy_solve_python(
    candidates: _typechecking.CandidatePairs
) -> _typechecking.MatchGroups:
    """Select matches from candidate pairs using the greedy algorithm.

    We assign each record to exactly one 'group' of records that are
    pairwise matched together. We iterate over `candidates` in order of
    decreasing similarity. When we encounter a pair of records (let s be
    their similarity) that do not already belong to the same group, we
    merge their groups iff (a) every pair of records is permitted to be
    matched, and (b) the similarity of every pair of records is above s.
    The latter requirement is only for tie-breaking: we prefer groups
    whose minimum pairwise similarity is highest. Any group merge
    rejected due to requirement (b) will be revisited later with that
    requirement relaxed.

    :param tuple candidates: Candidates, as returned by
        `find_candidates`.

    :return: An sequence of groups. Each group is an sequence of
        records. Two records are in the same group iff they represent
        the same entity. Here, a record is a two-tuple of dataset index
        and record index.
    """
    return probabilistic_greedy_solve_python(
        candidates, merge_threshold=1.0, deduplicated=False)



from bitarray import bitarray
import numpy as np

from anonlink.typechecking import FloatArrayType, IntArrayType


def sort_similarities_inplace(
    sims: FloatArrayType,
    rec_is0: IntArrayType,
    rec_is1: IntArrayType
) -> None:
    np_sims = np.frombuffer(sims, dtype=sims.typecode)
    np_indices0 = np.frombuffer(rec_is0, dtype=rec_is0.typecode)
    np_indices1 = np.frombuffer(rec_is1, dtype=rec_is1.typecode)
    
    np.negative(np_sims, out=np_sims)  # Sort in reverse.
    # Mergesort is stable. We need that for correct tiebreaking.
    order = np.argsort(np_sims, kind='mergesort')
    np.negative(np_sims, out=np_sims)

    # This modifies the original arrays since they share a buffer.
    np_sims[:] = np_sims[order]
    np_indices0[:] = np_indices0[order]
    np_indices1[:] = np_indices1[order]


def to_bitarray(record) -> bitarray:
    if isinstance(record, bitarray):
        return record
    if not isinstance(record, bytes):
        try:  # Not bitarray, not bytes. But is it bytes-like?
            record = bytes(record)
        except TypeError:
            raise TypeError('unsupported record type') from None
    ba = bitarray()
    ba.frombytes(record)
    return ba


def to_bitarrays(records):
    if all(isinstance(record, bitarray) for record in records):
        return records
    else:
        return tuple(map(to_bitarray, records))




cimport cython

from cpython cimport array
import array


from anonlink.similarities._dice cimport popcount_arrays as c_popcount_arrays
from anonlink.similarities._dice cimport dice_coeff as c_dice_coeff
from anonlink.similarities._dice cimport match_one_against_many_dice_k_top as c_match_one_against_many_dice_k_top

@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
cpdef unsigned int[::1] popcount_arrays(const char[::1] input_data, unsigned int array_bytes = 128):
    """
    Compute the popcount of a flattened array of data where each element
    is array_bytes long.
    """
    cdef array.array unsigned_int_array_template = array.array('I', [])
    cdef array.array output_counts
    if array_bytes == 0 or len(input_data) == 0:
        return array.clone(unsigned_int_array_template, 0, zero=True)
    else:
        output_size = len(input_data) // array_bytes
    # The CPP code reasonably assumes the length of the data is evenly divided by the array_bytes
    assert len(input_data) % array_bytes == 0, "input data length not divisible by array_bytes"
    output_counts = array.clone(unsigned_int_array_template, output_size, zero=True)
    popcount_arrays_preallocated_output(output_counts, input_data, array_bytes)
    return output_counts


@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
def popcount_arrays_preallocated_output(
        unsigned int[::1] output_counts,
        const char[::1] input_data,
        unsigned int array_bytes = 128
):
    """
    :param input_data: flattened contiguous char input of ARRAY_BYTES elements
    """
    cdef double elapsed_time

    # Create a memoryview of the input data and preallocated count results
    cdef const char[::1] arr_memview = input_data
    cdef unsigned int[::1] counts_memview = output_counts
    cdef unsigned int num_elements = <unsigned int>arr_memview.shape[0] // array_bytes

    with nogil:
        elapsed_time = c_popcount_arrays(
            &counts_memview[0],
            &arr_memview[0],
            num_elements,
            array_bytes)

    return elapsed_time


@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
def dice_coeff(
        const char[::1] input_data_1,
        const char[::1] input_data_2,
        int array_bytes = 128
):
    assert array_bytes % 8 == 0
    cdef const char[::1] memview_1 = input_data_1
    cdef const char[::1] memview_2 = input_data_2
    cdef double score

    with nogil:
        score = c_dice_coeff(
            &memview_1[0],
            &memview_2[0],
            array_bytes)

    return score


@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
cpdef int match_one_to_many_dice_preallocated_output(
        const char[::1] one,
        const char[::1] many,
        unsigned int[::1] counts_many,
        int n,
        int array_bytes,
        int k,
        double threshold,
        unsigned int[::1] output_indicies,
        double[::1] output_scores
) nogil:
    cdef int number_matched

    # Create a memoryview of the input data and preallocated results arrays
    cdef const char[::1] one_memview = one
    cdef const char[::1] many_memview = many
    cdef unsigned int[::1] counts_memview = counts_many
    cdef unsigned int[::1] indicies_memview = output_indicies
    cdef double[::1] scores_memview = output_scores

    cdef unsigned int num_elements = <unsigned int>many_memview.shape[0] // array_bytes

    number_matched = c_match_one_against_many_dice_k_top(
        &one_memview[0],
        &many_memview[0],
        &counts_memview[0],
        num_elements,
        array_bytes,
        k,
        threshold,
        &indicies_memview[0],
        &scores_memview[0]
    )

    return number_matched

@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)   # Deactivate negative indexing.
def dice_many_to_many(
        const char[::1] carr0,
        const char[::1] carr1,
        unsigned int length_f0,
        unsigned int length_f1,
        unsigned int[::1] c_popcounts,
        int filter_bytes,
        int k,
        double threshold,
        array.array result_sims,
        array.array result_indices0,
        array.array result_indices1
):
    cdef size_t i
    cdef int matches
    cdef int total_matches = 0
    assert len(carr1) == filter_bytes * length_f1
    assert len(c_popcounts) == length_f1

    # do all buffer allocations in Python and pass a memoryview to C
    cdef array.array double_array_template = array.array('d', [])
    cdef array.array int_array_template = array.array('I', [])

    cdef array.array c_scores
    cdef array.array c_indicies

    c_scores = array.clone(double_array_template, k, zero=True)
    c_indices = array.clone(int_array_template, k, zero=True)
    i_buffer = array.clone(int_array_template, k, zero=True)

    cdef double[::1] scores_memview = c_scores
    cdef unsigned int[::1] indicies_memview = c_indices
    cdef unsigned int[::1] i_buffer_memview = i_buffer

    for i in range(length_f0):
        matches = match_one_to_many_dice_preallocated_output(
            carr0[i * filter_bytes:(i + 1) * filter_bytes],
            carr1,
            c_popcounts,
            length_f1,
            filter_bytes,
            k,
            threshold,
            indicies_memview,
            scores_memview
        )
        total_matches += matches
        i_buffer_memview[:] = i

        assert matches <= k
        result_sims.extend(c_scores[:matches])
        result_indices0.extend(i_buffer[:matches])
        result_indices1.extend(c_indices[:matches])

    return total_matches
